/*
 * data.c - Values, variables, and operators
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <assert.h>

#include "dwarfx.h"

#include "util.h"
#include "data.h"


extern void yyerror(const char *s);

const struct dwx_type *void_type;
const struct dwx_type *int8_type;
const struct dwx_type *int16_type;
const struct dwx_type *int32_type;
const struct dwx_type *int64_type;
const struct dwx_type *uint8_type;
const struct dwx_type *uint16_type;
const struct dwx_type *uint32_type;
const struct dwx_type *uint64_type;

/*
 * Memory objects that identify types which are private to the script
 * processor, and don't exist in this form in C.
 */

struct dwx_type type_type;
struct dwx_type string_type;
struct dwx_type lvalue_type;
struct dwx_type array_type;


struct value int_value(uint32_t num)
{
    struct value value;

    value.type = uint32_type;
    /* @@@ memory leak. should reference count. */
    value.value = alloc_t(uint32_t);
    *(uint32_t *) value.value = num;
    return value;
}


uint32_t value_int(struct value value)
{
    if (!value.type) return 0;
    if (value.type->kind != dwx_uint && value.type->kind != dwx_sint)
        yyerror("invalid type (expected integer)");
    switch (value.type->size) {
	case 1:
	    return *(uint8_t *) value.value;
	case 2:
	    return *(uint16_t *) value.value;
	case 4:
	    return *(uint32_t *) value.value;
	case 8:
	    /* @@@ not yet supported */
	default:
	    abort();
    }
}


uint32_t int_eval(const struct op *op)
{
    return value_int(evaluate(op));
}


struct value string_value(const char *str)
{
    struct value value;

    value.type = &string_type;
    value.value = str;
    return value;
}


const char *value_string(struct value value)
{
    if (!value.type) return "";
    if (value.type != &string_type) yyerror("invalid type (expected string)");
    return value.value;
}


const char *string_eval(const struct op *op)
{
    return value_string(evaluate(op));
}


struct value type_value(const struct dwx_type *type)
{
    struct value value;

    value.type = &type_type;
    value.value = type;
    return value;
}


const struct dwx_type *value_type(struct value value)
{
    if (value.type != &type_type) yyerror("invalid type (expected type)");
    return value.value;
}


const struct dwx_type *type_eval(const struct op *op)
{
    return value_type(evaluate(op));
}


struct value array_value(void)
{
    struct value value;

    value.type = &array_type;
    value.value = NULL;
    return value;
}


struct value evaluate_lvalue(const struct op *op)
{
     return op->fn(op);
}


struct value dereference_lvalue(struct value value)
{
     assert(value.type == &lvalue_type);
     return *(struct value *) value.value;
}


struct value evaluate(const struct op *op)
{
     struct value value;

     value = op->fn(op);
     if (value.type != &lvalue_type) return value;
     return *(struct value *) value.value;
}


void init_types(void)
{
    int8_type = dwx_basic_type(dwx_sint,1);
    int16_type = dwx_basic_type(dwx_sint,2);
    int32_type = dwx_basic_type(dwx_sint,4);
    int64_type = dwx_basic_type(dwx_sint,8);
    uint8_type = dwx_basic_type(dwx_uint,1);
    uint16_type = dwx_basic_type(dwx_uint,2);
    uint32_type = dwx_basic_type(dwx_uint,4);
    uint64_type = dwx_basic_type(dwx_uint,8);
}
