/*
 * op.c - Operator construction and evaluation
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "dwarfx.h"

#include "util.h"
#include "data.h"
#include "array.h"
#include "io.h"
#include "op.h"


/* ----- Evaluation functions ---------------------------------------------- */


static struct value op_const(const struct op *self)
{
    return self->u.value;
}


static struct value op_get_var(const struct op *self)
{
    struct value value;

    value.type = &lvalue_type;
    value.value = &self->u.var->value;
    return value;
}


struct value op_logical_or(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) || int_eval(self->u.arg.b));
}


struct value op_logical_and(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) && int_eval(self->u.arg.b));
}


struct value op_bit_or(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) | int_eval(self->u.arg.b));
}


struct value op_bit_xor(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) ^ int_eval(self->u.arg.b));
}


struct value op_bit_and(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) & int_eval(self->u.arg.b));
}


struct value op_eq(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) == int_eval(self->u.arg.b));
}


struct value op_ne(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) != int_eval(self->u.arg.b));
}


struct value op_lt(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) < int_eval(self->u.arg.b));
}


struct value op_gt(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) > int_eval(self->u.arg.b));
}


struct value op_le(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) <=int_eval(self->u.arg.b));
}


struct value op_ge(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) >= int_eval(self->u.arg.b));
}


struct value op_shl(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) << int_eval(self->u.arg.b));
}


struct value op_shr(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) >> int_eval(self->u.arg.b));
}


struct value op_add(const struct op *self)
{
/* @@@ this is of course nonsense. should do proper type propagation */
    return int_value(int_eval(self->u.arg.a)+int_eval(self->u.arg.b));
}


struct value op_sub(const struct op *self)
{
    struct value a,b;
    int size;

    a = evaluate(self->u.arg.a);
    b = evaluate(self->u.arg.b);
    if (a.type->kind != dwx_ptr || b.type->kind != dwx_ptr)
	return int_value(value_int(a)-value_int(b));
    if (!dwx_type_equal(a.type,b.type)) {
	fprintf(stderr,"can only subtract identical types\n");
	exit(1);
    }
    if (!a.type->u.modifier.type) size = 1;
    else size = dwx_sizeof(a.type->u.modifier.type);
    return int_value((*(unsigned long *) a.value-*(unsigned long *) b.value)/
      size);
}


struct value op_mult(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a)*int_eval(self->u.arg.b));
}


struct value op_div(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a)/int_eval(self->u.arg.b));
}


struct value op_mod(const struct op *self)
{
    return int_value(int_eval(self->u.arg.a) % int_eval(self->u.arg.b));
}


struct value op_minus(const struct op *self)
{
    return int_value(-int_eval(self->u.arg.a));
}


struct value op_logical_not(const struct op *self)
{
    return int_value(!int_eval(self->u.arg.a));
}


struct value op_bit_not(const struct op *self)
{
    return int_value(~int_eval(self->u.arg.a));
}


/* ----- Constructors for constants ---------------------------------------- */


struct op *op_int(uint32_t num)
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = op_const;
    op->u.value = int_value(num);
    return op;
}


struct op *op_string(const char *str)
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = op_const;
    op->u.value = string_value(str);
    return op;
}


struct op *op_type(const struct dwx_type *type)
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = op_const;
    op->u.value = type_value(type);
    return op;
}


/* ----- Constructors for variables and types operators -------------------- */


struct op *op_var(const struct var *var)
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = op_get_var;
    op->u.var = var;
    return op;
}


static int size_of_value(struct value value)
{
    if (!value.type) {
	fprintf(stderr,"undefined values have no size\n");
	exit(1);
    }
    if (value.type == &string_type) return strlen(value.value)+1;
    if (value.type == &array_type) {
	fprintf(stderr,"invalid type\n");
	exit(1);
    }
    return dwx_sizeof(value.type);
}


struct value op_cast(const struct op *self)
{
    const struct dwx_type *type;
    struct value expr;

    type = type_eval(self->u.arg.a);
    expr = evaluate(self->u.arg.b);
    if (dwx_sizeof(type) > size_of_value(expr)) {
	fprintf(stderr,"cannot cast to a larger type\n");
	exit(1);
    }
    expr.type = type;
    return expr;
}


struct value op_sizeof_expr(const struct op *self)
{
    return int_value(size_of_value(evaluate(self->u.arg.a)));
}


struct value op_sizeof_type(const struct op *self)
{
    return int_value(dwx_sizeof(type_eval(self->u.arg.a)));
}


struct value op_index(const struct op *self)
{
    struct value value;
    unsigned long offset;

    value = evaluate_lvalue(self->u.arg.a);
    if (value.type == &lvalue_type) {
	if (!((struct value *) value.value)->type)
	    *(struct value *) value.value = array_value();
	if (((struct value *) value.value)->type == &array_type)
	    return index_array((struct value *) value.value,
	      evaluate(self->u.arg.b));
	value = dereference_lvalue(value);
    }
    if (value.type->kind != dwx_array) {
	fprintf(stderr,"trying to index a non-array\n");
	exit(1);
    }
    offset = (unsigned long) value.value;
    value.type = dwx_index(value.type,&offset,int_eval(self->u.arg.b));
    value.value = (const void *) offset;
    return value;
}


struct value op_member(const struct op *self)
{
    struct value value;
    unsigned long offset;

    value = evaluate(self->u.arg.a);
    if (value.type->kind != dwx_struct && value.type->kind != dwx_union) {
	fprintf(stderr,"trying to access member a non-struct/union\n");
	exit(1);
    }
    offset = (unsigned long) value.value;
    value.type = dwx_member(value.type,&offset,string_eval(self->u.arg.b));
    value.value = (const void *) offset;
    return value;
}


struct value op_add_array(const struct op *self)
{
    return type_value(
      dwx_add_array(type_eval(self->u.arg.a),int_eval(self->u.arg.b),-1));
}


struct value op_add_fn(const struct op *self)
{
    return type_value(dwx_add_fn(type_eval(self->u.arg.a)));
}


struct value op_add_ptr(const struct op *self)
{
    return type_value(dwx_add_ptr(type_eval(self->u.arg.a)));
}


/* ----- I/O operators ----------------------------------------------------- */


struct value op_read(const struct op *self)
{
    struct value value;
    const char *s,*p;
    int n;

    value = evaluate(self->u.arg.a);
    if (value.type == &type_type) {
	value.type = value_type(value);
	value.value = in_read(dwx_sizeof(value.type));
	return value;
    }
    s = value_string(value);
    n = strlen(s)+1;
    p = in_read_n(n);
    if (!p) return int_value(0);
    if (strncmp(p,s,n)) {
	in_return(p,n);
	return int_value(0);
    }
    in_free(p);
    return int_value(1);
}


struct value op_more_data(const struct op *self)
{
    return int_value(!in_eof());
}


/* ----- Conditional operator ---------------------------------------------- */


struct value op_conditional(const struct op *self)
{
    return evaluate_lvalue(int_eval(self->u.arg.a) ? self->u.arg.b :
      self->u.arg.c);
}


/* ----- Assignment operator ----------------------------------------------- */


struct value op_assignment(const struct op *self)
{
    struct value lvalue,rvalue;

    lvalue = evaluate_lvalue(self->u.arg.a);
    if (lvalue.type != &lvalue_type) {
        fprintf(stderr,"lvalue expected\n");
        exit(1);
    }
    /* @@@ memory leak: doesn't remove previous value */
    rvalue = evaluate(self->u.arg.b);
    *(struct value *) lvalue.value = copy_if_array(rvalue);
    return rvalue;
}


/* ----- Constructors for regular operators -------------------------------- */


struct op *op_noop(struct value (*fn)(const struct op *self))
{
    struct op *op;

    op = alloc_t(struct op);
    op->fn = fn;
    return op;
}


struct op *op_unary(struct value (*fn)(const struct op *self),
  const struct op *a)
{
    struct op *op;

    op = op_noop(fn);
    op->u.arg.a = a;
    return op;
}


struct op *op_binary(struct value (*fn)(const struct op *self),
  const struct op *a,const struct op *b)
{
    struct op *op;

    op = op_unary(fn,a);
    op->u.arg.b = b;
    return op;
}


struct op *op_ternary(struct value (*fn)(const struct op *self),
  const struct op *a,const struct op *b,const struct op *c)
{
    struct op *op;

    op = op_binary(fn,a,b);
    op->u.arg.c = c;
    return op;
}
