#!/bin/sh
#
# runtests.sh - Run regression tests from standard input
#
# Written 2001,2002 by Werner Almesberger
# Copyright 2001 EPFL-ICA
# Copyright 2001,2002 Bivio Networks, Network Robots
# Copyright 2002 Werner Almesberger, California Institute of Technology
#

#
# runtests.sh originates from the tcng package (http://tcng.sourceforge.net/),
# and has been modified for the requirements of dumpbb
#

#
# The input file(s) has/have the following structure:
#
# FILE <file_name>
# # <test_title>
# <command>
# ...
# INPUT
# <input>
# ...
# OUTPUT
# ERROR
# <output>
# ...
# CLEANUP
# <command>
# ...
# (EOF or next test)
#
# The directives FILE and ERROR are optional. The sections INPUT, OUTPUT, and
# CLEANUP contain zero or more lines, and can also be omitted. <command> is any
# valid shell command line (which may contain multiple commands), and may be
# spread over multiple lines by continuing lines with a backslash.
#
# The only comments allowed are the test titles, but <input> may contain lines
# beginning with a hash sign (#). runtests.sh defines the command dumpbb, so
# that no path names have to be specified to access it. It also provides the
# PID of the runtests.sh process in the environment variable PID ($$ changes
# between the test commands and the cleanup section).
#
# runtests.sh executes each command with standard input from the <input>
# section. Normally, all commands are expected to return with exit status zero
# (success), and to produce exactly the output defined in the <output> section.
# If the command writes anything to standard error, runtests.sh issues a
# warning, but does not count this as a problem.
#
# If the ERROR directive is present, the last command is expected to exit with
# non-zero exit status, and runtests.sh only issues a warning if the command's
# output does not correspond exactly to the <output> section. (Rationale: in
# this case, the output is probably an error message, and error messages are
# frequently changes or improved, so a difference does not necessarily indicate
# a problem.)
#
# The FILE directive indicates from which file the following tests originate,
# and is printed when runtests.sh encounters a problem.
#

#
# Test cases can also be conditionally executed (e.g. is using a subsystem that
# may not be available in all possible installations). In this case, the
# following structure is used:
#
# BEGIN CONDITIONAL
# test case without title
# conditional test case(s)
# ...
# END CONDITIONAL
#
# Note that the word "CONDITIONAL" must be immediately followed by a newline -
# trailing spaces are now allowed. Conditional tests cannot be nested.
#

[ -z "$TOPDIR" ] && TOPDIR=.


find_in_path()
{
    for n in `echo $2 | tr : ' '`; do
	if [ -x $n/$1 ]; then
	    echo $n/$1
	    return
	fi
    done
    echo "$1 not found" 1>&2
    exit 1
}


DUMPBB_CMD="`find_in_path dumpbb $TOPDIR/dumpbb:$PATH`"
VALGRIND=

set -a
[ -r config ] && . ./config
set +a
set -o noglob

usage()
{
    echo "usage: $0 [-d] [-g] [-i] [-u] [file ...]" 1>&2
    echo "  -d           print test output, don't compare with reference" 1>&2
    echo "  -g           use Valgrind" 1>&2
    echo "  -i           continue after errors" 1>&2
    echo "  -u           print output mismatches as unified context diffs" 1>&2
    exit 1
}


debug=false
stop_on_error=true
diff=false

while [ ! -z "$1" ]; do
    case "$1" in
	-c) count=true;;
	-d) debug=true;;
	-g) VALGRIND="valgrind -q --trace-children=yes"
	    VALGRIND="$VALGRIND --suppressions=$TOPDIR/build/valgrind.supp";;
	-i) stop_on_error=false;;
	-u) diff=true;;
	-*) usage;;
	*)  break;;
    esac
    shift
done

if [ -z "$TCNG_TESTS_COUNT" ]; then
    count=false
else
    count=true
fi

for n in "$@"; do
    <$n || exit 1
done


stderr()
{
    if [ -s _err.$$ ]; then
	echo "  Standard error:"
	sed 's/^/  | /' <_err.$$
    fi
    rm -f _err.$$
}


stdout()
{
    echo "  Standard output:"
    sed 's/^/  | /' <_out.$$
}


refout()
{
    echo "  Reference output:"
    sed 's/^/  | /' <_ref.$$
}


failed()
{
    echo "$1"
    stderr
    [ -z "$file" ] || echo "File:      $file"
    [ -s _cmd.$$ ] && echo "Commands:  _cmd.$$"
    [ -s _in.$$ ] && echo "Input:     _in.$$"
    [ -s _out.$$ ] && echo "Output:    _out.$$"
    [ -s _ref.$$ ] && echo "Reference: _ref.$$"
    [ -s _wipe.$$ ] && echo "Cleanup:   _wipe.$$"
    if $diff; then
	echo "----- diff -u _ref.$$ _out.$$ -----"
	diff -u _ref.$$ _out.$$
    fi
    $stop_on_error && exit 1
    echo
    failed=`expr $failed + 1`
}

first_file()
{
    file=`echo "$l" | sed '/^FILE /s///p;d'`
    if [ -z "$file" ]; then
	file="$next_file"
    else
	read l
    fi
    next_file="$file"
}


flush_cmd()
{
    if [ $part = cmd ]; then
	echo "if $cmd; then echo 0; else echo \$?; fi" >>_cmd.$$
    fi
}


#
# Input/output;
#   Variable:
#     l			next line in input
#     next_file		next file name (empty if unknown)
#
# Output:
#   Variables:
#     file		name of test case file (empty if unknown)
#     next_file		next file name (empty if unknown)
#     title		test title
#     cmd		test command
#     expect_error	true if test succeeds if command fails
#
#   Files:
#     _in.$$		standard input for test
#     _ref.$$		expected test output
#

read_test()
{
    first_file
    title="`echo "$l" | sed 's/# //;s/ -*$//'`"
    read cmd
    echo 'dumpbb() { '$DUMPBB_CMD' "$@"; }' >_cmd.$$
#    echo 'gcc() { gcc296 "$@"; }' >>_cmd.$$
    >_ref.$$
    >_in.$$
    >_wipe.$$
    part=cmd
    expect_error=false
    while read l; do
	echo "$l" | grep -s '^\(# \|BEGIN CONDITIONAL$\|END CONDITIONAL$\)' \
	  >/dev/null && break
	if [ "$l" = ERROR ]; then
	    expect_error=true
	elif [ "$l" = INPUT ]; then
	    flush_cmd
	    part=in
	elif [ "$l" = OUTPUT ]; then
	    flush_cmd
	    part=ref
	elif [ "$l" = CLEANUP ]; then
	    flush_cmd
	    part=wipe
	else
	    _file=`echo "$l" | sed '/^FILE /s///p;d'`
	    if [ -z "$_file" ]; then
		if [ $part = cmd ]; then
		    echo "$cmd" >>_cmd.$$
		    cmd="$l"
		else
		    echo "$l" >>_$part.$$
		fi
	    else
		next_file="$_file"
	    fi
	fi
    done
    flush_cmd
}


eval_success()
{
    result=`tail -1 _tmp_out.$$`
    sed -e '1{x;d;};x' <_tmp_out.$$ >_out.$$
#    sed 's/ *$//' <_tmp_out.$$ >_out.$$
    rm -f _tmp_out.$$
    cmp -s _out.$$ _ref.$$
    equal=$?
    if $expect_error; then
	if [ $result = 0 ]; then
	    if $condition; then
		skipping=true
	    else
		failed "FAILED (expected error)"
	    fi
	else
	    if [ $equal = 0 ]; then
		if $condition; then
		    skipping=false
		else
		    echo PASSED
		    rm -f _err.$$
		    c=`expr $c + 1`
		fi
	    else
		if $condition; then
		    echo "Warning: output differs in condition"
		    skipping=false
		else
		    echo "PASSED (warning: output differs)"
		    c=`expr $c + 1`
		fi
		stderr
		stdout
		refout
		w=`expr $w + 1`
	    fi
	fi
    else
	if [ $result != 0 ]; then
	    if $condition; then
		skipping=true
	    else
		failed "FAILED (got error)"
	    fi
	else
	    if [ $equal = 0 ]; then
		if $condition; then
		    skipping=false
		else
		    echo PASSED
		    stderr
		    c=`expr $c + 1`
		fi
	    else
		if $condition; then
		    skipping=true
		else
		    failed FAILED
		fi
	    fi
	fi
    fi
}


run_tests()
{
    conditional=false
    condition=false
    skipping=false
    c=0
    w=0
    skipped=0
    read l
    first_file
    while [ ! -z "$l" ]; do
	if [ "$l" = "BEGIN CONDITIONAL" ]; then
	    if $conditional; then
		echo "BEGIN CONDITIONAL in conditional block" 1>&2
		exit 1
	    fi
	    conditional=true
	    condition=true
	elif [ "$l" = "END CONDITIONAL" ]; then
	    if $conditional; then
		conditional=false
		skipping=false
	    else
		echo "END CONDITIONAL without condition" 1>&2
		exit 1
	    fi
	    read l
	    first_file
	    continue
	fi
	read_test
	if $skipping; then
	    skipped=`expr $skipped + 1`
	    rm -f _in.$$ _ref.$$
	    continue
	fi
	if $condition; then :; else
	    echo -n "$title: "
	fi
	if PID=$$ sh -e _cmd.$$ <_in.$$ >_tmp_out.$$ 2>_err.$$; then
	    eval_success
	else
	    failed "FAILED (early error)"
	fi
	if [ -s _wipe.$$ ]; then
	    PID=$$ sh -e _wipe.$$ || failed "FAILED (in cleanup)"
	fi
	rm -f _cmd.$$ _in.$$ _out.$$ _ref.$$ _wipe.$$
	condition=false
    done
    if $conditional; then
	echo "EOF in conditional block" 1>&2
	exit 1
    fi
    if [ "$c" = 1 ]; then
	echo -n "Passed the test"
    else
	echo -n "Passed all $c tests"
    fi
    if [ `expr $w  + $skipped` = 0 ]; then
	echo
    else
	echo -n " ("
	if [ $w != 0 ]; then
	    if [ $w = 1 ]; then
		echo -n "1 warning"
	    else
		echo -n "$w warnings"
	    fi
	    if [ $skipped != 0 ]; then
		echo -n ", "
	    fi
	fi
	if [ $skipped != 0 ]; then
	    if [ $skipped = 1 ]; then
		echo -n "1 conditional test skipped"
	    else
		echo -n "$skipped conditional tests skipped"
	    fi
	fi
	echo ")"
    fi
    if [ $failed -gt 0 ]; then
	echo "$failed test(s) FAILED !"
    fi
}


fill()
{
    # superbly inefficient, but hey, CPU cycles are cheap :-)
    s="$2$1$2"
    while [ `echo "$s" | wc -c` -lt 79 ]; do
	s="$2$s$2"
    done
    echo $s
}


debug_tests()
{
    read l
    while [ ! -z "$l" ]; do
	read_test
	fill " $title " =
	echo "$cmd"
	fill "" -
	eval "$cmd" <_in.$$
	rm -f _in.$$ _ref.$$
	echo
    done
}


count_tests()
{
    c=0
    read l
    while [ ! -z "$l" ]; do
	read_test
	rm -f _in.$$ _ref.$$
	c=`expr $c + 1`
	echo -en "$c\r"
    done
    echo "Counted $c test(s)"
}


cat_files()
{
    if [ -z "$*" ]; then
	cat
    else
	for n in "$@"; do
	    echo FILE "$n"
	    cat "$n"
	done
    fi
}


failed=0
next_file=

if $count; then
    cat_files "$@" | count_tests
elif $debug; then
    cat_files "$@" | debug_tests
else
    cat_files "$@" | run_tests
fi
