%{
/*
 * script.y - Dump script language
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */

/*
 * This grammar is loosely based on the ANSI C grammar posted by Jeff Lee in
 * 1985. ftp://ftp.uu.net/usenet/net.sources/ansi.c.grammar.Z
 */

/*
 * The differences between this scripting language and C are described in
 * README.
 */


#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>

#include "dwarfx.h"

#include "util.h"
#include "data.h"
#include "var.h"
#include "op.h"
#include "stmt.h"


extern int lineno;

struct stmt *script;
static struct op *curr_type; /* type under construction */


static struct while_loop {
    const struct stmt *loop;
    const char *label;
    struct while_loop *next;
} *while_stack = NULL;


static void push_while(const struct stmt *loop,const char *label)
{
    struct while_loop *entry;

    entry = alloc_t(struct while_loop);
    entry->loop = loop;
    entry->label = label;
    entry->next = while_stack;
    while_stack = entry;
}


static void pop_while(void)
{
    struct while_loop *next;

    next = while_stack->next;
    if (while_stack->label) free(while_stack->label);
    free(while_stack);
    while_stack = next;
}


%}


%union {
    uint32_t num;
    const char *str;
    struct var *var;
    const struct op *op;
    struct stmt *stmt;
    const struct without *without;
    const struct arg_list *args;
};


%token		TOK_UNSIGNED TOK_SIGNED
%token		TOK_VOID TOK_CHAR TOK_SHORT TOK_INT TOK_LONG
%token		TOK_INT8_T TOK_INT16_T TOK_INT32_T TOK_INT64_T
%token		TOK_UINT8_T TOK_UINT16_T TOK_UINT32_T TOK_UINT64_T
%token		TOK_STRUCT TOK_UNION TOK_ENUM

%token		TOK_PRINT TOK_PRINTF TOK_WITHOUT TOK_READ
%token		TOK_IF TOK_ELSE TOK_WHILE TOK_BREAK TOK_EXIT TOK_DIE

%token		TOK_OR TOK_AND TOK_SHL TOK_SHR TOK_EQ TOK_NE TOK_LE TOK_GE
%token		TOK_SIZEOF

%token <num>	NUMBER ENUM
%token <str>	WORD STRING
%token <var>	VARIABLE

%type <str>	opt_label_name opt_label

%type <op>	expression opt_expression
%type <op>	assignment_expression conditional_expression
%type <op>	logical_or_expression logical_and_expression
%type <op>	inclusive_or_expression exclusive_or_expression
%type <op>	and_expression
%type <op>	equality_expression relational_expression shift_expression
%type <op>	additive_expression multiplicative_expression
%type <op>	cast_expression unary_expression
%type <op>	postfix_expression primary_expression

%type <op>	type_name type_specifier type_name_or_expression

%type <stmt>	statements statement
%type <stmt>	block expression_statement print printf
%type <stmt>	if opt_else while break exit die

%type <without>	opt_without without_list without

%type <args>	arg_list arg opt_arg_list

%nonassoc BENEATH_ELSE
%nonassoc TOK_ELSE

%%


/* ===== Statements and blocks ============================================= */


all:
    statements
	{
	    script = $1;
	    fix_statements(&script);
	}
    ;

statements:
	{
	    $$ = NULL;
	}
    | statement statements
	{
	    if (!$1) $$ = $2;
	    else {
		$$ = $1;
		$$->next = $2;
	    }
	}
    ;

statement:
    block
	{
	    $$ = $1;
	}
    | expression_statement
	{
	    $$ = $1;
	}
    | print
	{
	    $$ = $1;
	}
    | printf
	{
	    $$ = $1;
	}
    | if
	{
	    $$ = $1;
	}
    | while
	{
	    $$ = $1;
	}
    | break
	{
	    $$ = $1;
	}
    | exit
	{
	    $$ = $1;
	}
    | die
	{
	    $$ = $1;
	}
    | ';'
	{
	    $$ = NULL;
	}
    ;

block:
    '{' statements '}'
	{
	    $$ = $2;
	}
    ;


/* ----- Assignment -------------------------------------------------------- */


expression_statement:
    expression ';'
	{
	    $$ = alloc_t(struct stmt);
	    $$->fn = stmt_expression;
	    $$->u.expr.expr = $1;
	}
    ;


/* ----- print ------------------------------------------------------------- */


print:
    TOK_PRINT arg_list opt_without ';'
	{
	    $$ = alloc_t(struct stmt);
	    $$->fn = stmt_print;
	    $$->u.print.args = $2;
	    $$->u.print.without = $3;
	}
    ;

arg_list:
    arg
	{
	    $$ = $1;
	}
    | arg ',' arg_list
	{
	    $$ = $1;
	    $$->next = $3;
	}
    ;

arg:
    expression
	{
	    $$ = alloc_t(struct arg_list);
	    $$->expr = $1;
	    $$->next = NULL;
	}
    ;

opt_without:
	{
	    $$ = NULL;
	}
    | TOK_WITHOUT without_list
	{
	    $$ = $2;
	}
    ;

without_list:
    without
	{
	    $$ = $1;
	}
    | without ',' without_list
	{
	    $$ = $1;
	    $$->next = $3;
	}
    ;

without:
    expression
	{
	    $$ = alloc_t(struct without);
	    $$->expr = $1;
	    $$->next = NULL;
	}
    ;


/* ----- printf ------------------------------------------------------------ */


printf:
    TOK_PRINTF '(' expression opt_arg_list ')' ';'
	{
	    $$ = alloc_t(struct stmt);
	    $$->fn = stmt_printf;
	    $$->u.printf.fmt = $3;
	    $$->u.printf.args = $4;
	}
    ;

opt_arg_list:
	{
	    $$ = NULL;
	}
    | ',' arg_list
	{
	    $$ = $2;
	}
    ;


/* ----- if ---------------------------------------------------------------- */


if:
    TOK_IF '(' expression ')' statement opt_else
	{
	    $$ = alloc_t(struct stmt);
	    $$->fn = stmt_if;
	    $$->u.if_stmt.expr = $3;
	    $$->u.if_stmt.true = $5;
	    $$->u.if_stmt.false = $6;
	}
    ;


opt_else:
    %prec BENEATH_ELSE
	{
	    $$ = NULL;
	}
    | TOK_ELSE statement
	{
	    $$ = $2;
	}
    ;

/* ----- break ------------------------------------------------------------- */


break:
    TOK_BREAK opt_label_name ';'
	{
	    $$ = alloc_t(struct stmt);
	    $$->fn = stmt_break;
	    if (!$2) {
		if (!while_stack) yyerror("break without while");
		$$->u.break_stmt.loop = while_stack->loop;
	    }
	    else {
		const struct while_loop *walk;

		for (walk = while_stack; walk; walk = walk->next)
		    if (walk->label && !strcmp(walk->label,$2)) break;
		if (!walk) yyerror("no such label");
		$$->u.break_stmt.loop = walk->loop;
		free($2);
	    }
	}
    ;

opt_label_name:
	{
	    $$ = NULL;
	}
    | WORD
	{
	    $$ = $1;
	}
    ;


/* ----- exit -------------------------------------------------------------- */


exit:
    TOK_EXIT opt_expression ';'
	{
	    $$ = alloc_t(struct stmt);
	    $$->fn = stmt_exit;
	    $$->u.exit.expr = $2;
	}
    ;

opt_expression:
	{
	    $$ = NULL;
	}
    | expression
	{
	    $$ = $1;
	}
    ;


/* ----- die -------------------------------------------------------------- */


die:
    TOK_DIE opt_expression ';'
	{
	    $$ = alloc_t(struct stmt);
	    $$->fn = stmt_die;
	    $$->u.die.expr = $2;
	    $$->u.die.line = lineno;
	}
    ;


/* ----- while ------------------------------------------------------------- */


while:
    opt_label TOK_WHILE '(' expression ')'
	{
	    $$ = alloc_t(struct stmt);
	    $$->fn = stmt_while;
	    push_while($$,$1);
	}
      statement
	{
	    pop_while();
	    $$ = $<stmt>6;
	    $$->u.while_stmt.expr = $4;
	    $$->u.while_stmt.body = $7;
	}
    ;


opt_label:
	{
	    $$ = NULL;
	}
    | WORD ':'
	{
	    $$ = $1;
	}
    ;

/* ===== Types ============================================================= */


type_name:
    type_specifier
	{
	    $$ = $1;
	}
    | type_specifier
	{
	    curr_type = $1;
	}
       abstract_declarator
	{
	    $$ = curr_type;
	}
    ;


/* ----- Basic types ------------------------------------------------------- */


type_specifier:
    TOK_VOID
	{
	    $$ = op_type(NULL);
	}
    | TOK_CHAR
	{
	    $$ = op_type(int8_type);
	}
    | TOK_SIGNED TOK_CHAR
	{
	    $$ = op_type(int8_type);
	}
    | TOK_UNSIGNED TOK_CHAR
	{
	    $$ = op_type(uint8_type);
	}
    | TOK_SHORT
	{
	    $$ = op_type(int16_type);
	}
    | TOK_SIGNED TOK_SHORT
	{
	    $$ = op_type(int16_type);
	}
    | TOK_UNSIGNED TOK_SHORT
	{
	    $$ = op_type(uint16_type);
	}
    | TOK_INT
	{
	    $$ = op_type(int32_type);
	}
    | TOK_SIGNED TOK_INT
	{
	    $$ = op_type(int32_type);
	}
    | TOK_UNSIGNED TOK_INT
	{
	    $$ = op_type(uint32_type);
	}
    | TOK_LONG
	{
	    $$ = op_type(int32_type);
	}
    | TOK_SIGNED TOK_LONG
	{
	    $$ = op_type(int32_type);
	}
    | TOK_UNSIGNED TOK_LONG
	{
	    $$ = op_type(uint32_type);
	}
    | TOK_LONG TOK_LONG
	{
	    $$ = op_type(int64_type);
	}
    | TOK_SIGNED TOK_LONG TOK_LONG
	{
	    $$ = op_type(int64_type);
	}
    | TOK_UNSIGNED TOK_LONG TOK_LONG
	{
	    $$ = op_type(uint64_type);
	}
    | TOK_INT8_T
	{
	    $$ = op_type(int8_type);
	}
    | TOK_INT16_T
	{
	    $$ = op_type(int16_type);
	}
    | TOK_INT32_T
	{
	    $$ = op_type(int32_type);
	}
    | TOK_INT64_T
	{
	    $$ = op_type(int64_type);
	}
    | TOK_UINT8_T
	{
	    $$ = op_type(uint8_type);
	}
    | TOK_UINT16_T
	{
	    $$ = op_type(uint16_type);
	}
    | TOK_UINT32_T
	{
	    $$ = op_type(uint32_type);
	}
    | TOK_UINT64_T
	{
	    $$ = op_type(uint64_type);
	}
    | TOK_STRUCT WORD
	{
	    const struct dwx_type *type;

	    type = dwx_lookup_type(dwx_struct,$2,-1,0);
	    if (!type) yyerror("unknown struct");
	    free($2);
	    $$ = op_type(type);
	}
    | TOK_UNION WORD
	{
	    const struct dwx_type *type;

	    type = dwx_lookup_type(dwx_union,$2,-1,0);
	    if (!type) yyerror("unknown union");
	    free($2);
	    $$ = op_type(type);
	}
    | TOK_ENUM WORD
	{
	    const struct dwx_type *type;

	    type = dwx_lookup_type(dwx_enum,$2,-1,0);
	    if (!type) yyerror("unknown enum");
	    free($2);
	    $$ = op_type(type);
	}
    | WORD
	{
	    const struct dwx_type *type;

	    type = dwx_lookup_type(dwx_typedef,$1,-1,0);
	    if (!type) yyerror("unknown typedef");
	    free($1);
	    $$ = op_type(type);
	}
    ;


/* ----- Abstract declarators ---------------------------------------------- */


abstract_declarator:
    pointer
    | abstract_declarator2
    | pointer abstract_declarator2
    ;

abstract_declarator2:
    '(' abstract_declarator ')'
    | '[' expression ']'
	{
	    curr_type = op_binary(op_add_array,curr_type,$2);
	}
    | abstract_declarator2 '[' expression ']'
	{
	    curr_type = op_binary(op_add_array,curr_type,$3);
	}
    | '(' ')'
	{
	    curr_type = op_unary(op_add_fn,curr_type);
	}
    | abstract_declarator2 '(' ')'
	{
	    curr_type = op_unary(op_add_fn,curr_type);
	}
    ;

pointer:
    '*'
	{
	    curr_type = op_unary(op_add_ptr,curr_type);
	}
    | '*' pointer
	{
	    curr_type = op_unary(op_add_ptr,curr_type);
	}
    ;


/* ===== Expressions ======================================================= */


expression:
    assignment_expression
	{
	    $$ = $1;
	}
    ;

assignment_expression:
    conditional_expression
	{
	    $$ = $1;
	}
    | unary_expression '=' assignment_expression
	{
	    $$ = op_binary(op_assignment,$1,$3);
	}
    ;

conditional_expression:
    logical_or_expression
	{
	    $$ = $1;
	}
    | logical_or_expression '?' logical_or_expression ':'
      conditional_expression
	{
	    $$ = op_ternary(op_conditional,$1,$3,$5);
	}
    ;

logical_or_expression:
    logical_and_expression
	{
	    $$ = $1;
	}
    | logical_or_expression TOK_OR logical_and_expression
	{
	    $$ = op_binary(op_logical_or,$1,$3);
	}
    ;

logical_and_expression:
    inclusive_or_expression
	{
	    $$ = $1;
	}
    | logical_and_expression TOK_AND inclusive_or_expression
	{
	    $$ = op_binary(op_logical_and,$1,$3);
	}
    ;

inclusive_or_expression:
    exclusive_or_expression
	{
	    $$ = $1;
	}
    | inclusive_or_expression '|' exclusive_or_expression
	{
	    $$ = op_binary(op_bit_or,$1,$3);
	}
    ;

exclusive_or_expression:
    and_expression
	{
	    $$ = $1;
	}
    | exclusive_or_expression '^' and_expression
	{
	    $$ = op_binary(op_bit_xor,$1,$3);
	}
    ;

and_expression:
    equality_expression
	{
	    $$ = $1;
	}
    | and_expression '&' equality_expression
	{
	    $$ = op_binary(op_bit_and,$1,$3);
	}
    ;

equality_expression:
    relational_expression
	{
	    $$ = $1;
	}
    | equality_expression TOK_EQ relational_expression
	{
	    $$ = op_binary(op_eq,$1,$3);
	}
    | equality_expression TOK_NE relational_expression
	{
	    $$ = op_binary(op_ne,$1,$3);
	}
    ;

relational_expression:
    shift_expression
	{
	    $$ = $1;
	}
    | relational_expression '<' shift_expression
	{
	    $$ = op_binary(op_lt,$1,$3);
	}
    | relational_expression '>' shift_expression
	{
	    $$ = op_binary(op_gt,$1,$3);
	}
    | relational_expression TOK_LE shift_expression
	{
	    $$ = op_binary(op_le,$1,$3);
	}
    | relational_expression TOK_GE shift_expression
	{
	    $$ = op_binary(op_ge,$1,$3);
	}
    ;

shift_expression:
    additive_expression
	{
	    $$ = $1;
	}
    | shift_expression TOK_SHL additive_expression
	{
	    $$ = op_binary(op_shl,$1,$3);
	}
    | shift_expression TOK_SHR additive_expression
	{
	    $$ = op_binary(op_shr,$1,$3);
	}
    ;

additive_expression:
    multiplicative_expression
	{
	    $$ = $1;
	}
    | additive_expression '+' multiplicative_expression
	{
	    $$ = op_binary(op_add,$1,$3);
	}
    | additive_expression '-' multiplicative_expression
	{
	    $$ = op_binary(op_sub,$1,$3);
	}
    ;

multiplicative_expression:
    cast_expression
	{
	    $$ = $1;
	}
    | multiplicative_expression '*' cast_expression
	{
	    $$ = op_binary(op_mult,$1,$3);
	}
    | multiplicative_expression '/' cast_expression
	{
	    $$ = op_binary(op_div,$1,$3);
	}
    | multiplicative_expression '%' cast_expression
	{
	    $$ = op_binary(op_mod,$1,$3);
	}
    ;

cast_expression:
    unary_expression
	{
	    $$ = $1;
	}
    | '(' type_name ')' cast_expression
	{
	    $$ = op_binary(op_cast,$2,$4);
	}
    ;

unary_expression:
    postfix_expression
	{
	    $$ = $1;
	}
    | '-' cast_expression
	{
	    $$ = op_unary(op_minus,$2);
	}
    | '+' cast_expression
	{
	    $$ = $2;
	}
    | '!' cast_expression
	{
	    $$ = op_unary(op_logical_not,$2);
	}
    | '~' cast_expression
	{
	    $$ = op_unary(op_bit_not,$2);
	}
    | TOK_SIZEOF unary_expression
	{
	    $$ = op_unary(op_sizeof_expr,$2);
	}
    | TOK_SIZEOF '(' type_name ')'
	{
	    $$ = op_unary(op_sizeof_type,$3);
	}
    ;

postfix_expression:
    primary_expression
	{
	    $$ = $1;
	}
    | postfix_expression '[' expression ']'
	{
	    $$ = op_binary(op_index,$1,$3);
	}
    | postfix_expression '.' WORD
	{
	    $$ = op_binary(op_member,$1,op_string($3));
	}
    | TOK_READ '(' type_name_or_expression ')'
	{
	    $$ = op_unary(op_read,$3);
	}
    | TOK_READ
	{
	    $$ = op_noop(op_more_data);
	}
    ;

type_name_or_expression:
    type_name
	{
	    $$ = $1;
	}
    | expression
	{
	    $$ = $1;
	}
    ;

primary_expression:
    VARIABLE
	{
	    $$ = op_var($1);
	}
    | NUMBER
	{
	    $$ = op_int($1);
	}
    | ENUM
	{
	    $$ = op_int($1);
	}
    | STRING
	{
	    $$ = op_string($1);
	}
    | '(' expression ')'
	{
	    $$ = $2;
	}
    ;
