/*
 * stmt.c - Statement construction and execution
 *
 * Written 2002,2003 by Werner Almesberger, Caltech Netlab FAST project
 */


#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "dwarfx.h"

#include "data.h"
#include "array.h"
#include "io.h"
#include "sprintf_generic.h"
#include "stmt.h"


static const struct stmt *ip = NULL;


/* ----- Fix statement pointers -------------------------------------------- */


static void end_sequence(struct stmt **stmt,struct stmt *next)
{
    int more;

    if (!*stmt) {
	*stmt = next;
	return;
    }
    more = 1;
    while (more) {
	if (!(*stmt)->next) {
	    (*stmt)->next = next;
	    more = 0;
	}
	if ((*stmt)->fn == stmt_if) {
	    end_sequence(&(*stmt)->u.if_stmt.true,(*stmt)->next);
	    end_sequence(&(*stmt)->u.if_stmt.false,(*stmt)->next);
	}
	else if ((*stmt)->fn == stmt_while)
	    end_sequence(&(*stmt)->u.while_stmt.body,*stmt);
	else if ((*stmt)->fn == stmt_break)
	    (*stmt)->u.break_stmt.loop = (*stmt)->u.break_stmt.loop->next;
	stmt = &(*stmt)->next;
    }
}


void fix_statements(struct stmt **stmt)
{
    end_sequence(stmt,NULL);
}


/* ----- Statement execution ----------------------------------------------- */


void stmt_expression(const struct stmt *self)
{
    /* @@@ memory leak */
    evaluate_lvalue(self->u.expr.expr);
}


void stmt_print(const struct stmt *self)
{
    const struct without *wo;
    const struct arg_list *arg;

    for (wo = self->u.print.without; wo; wo = wo->next) {
	struct value value;

	value = evaluate(wo->expr);
	dwx_output_skip(value.value,dwx_sizeof(value.type));
    }
    for (arg = self->u.print.args; arg; arg = arg->next) {
	struct value value;

	value = evaluate(arg->expr);
	if (value.type == &string_type)
	    printf("%s\n",(const char *) value.value);
	else dwx_output_data(stdout,value.type,value.value);
    }
    dwx_output_all();
}


#if 0
void stmt_read(const struct stmt *self)
{
    struct value value;
    const char *s;
    int n;
    const void *p;

    value = evaluate(self->u.read.expr);
    if (value.type == &type_type) {
	in_skip(dwx_sizeof(value_type(value)));
	return;
    }
    s = value_string(value);
    n = strlen(s)+1;
    p = in_read(n);
    if (strcmp(s,p)) {
	fprintf(stderr,"read mismatch\n");
	exit(1);
    }
    in_free(p);
}
#endif


void stmt_if(const struct stmt *self)
{
    ip = int_eval(self->u.if_stmt.expr) ?
      self->u.if_stmt.true : self->u.if_stmt.false;
}


void stmt_break(const struct stmt *self)
{
    ip = self->u.break_stmt.loop;
}


void stmt_exit(const struct stmt *self)
{
    if (!in_eof()) fprintf(stderr,"warning: unread dump data\n");
    exit(self->u.exit.expr ? int_eval(self->u.exit.expr) : 0);
}


void stmt_die(const struct stmt *self)
{
    if (self->u.die.expr) fprintf(stderr,"%s\n",string_eval(self->u.die.expr));
    else fprintf(stderr,"Died at line %u\n",self->u.die.line);
    exit(1);
}


void stmt_while(const struct stmt *self)
{
    if (int_eval(self->u.while_stmt.expr)) ip = self->u.while_stmt.body;
}


void execute(const struct stmt *pgm)
{
    ip = pgm;
    while (ip) {
	const struct stmt *stmt = ip;

//fprintf(stderr,"%p (%p)\n",ip,ip->fn);
	ip = ip->next;
	stmt->fn(stmt);
    }
}


/* ----- printf ------------------------------------------------------------ */


struct dsc {
    struct arg_list *curr;
    struct arg_list *list;
};


static void first_arg(void *dsc)
{
    struct dsc *d = dsc;

    d->curr = d->list;
}


static void next_arg(void *dsc)
{
    struct dsc *d = dsc;

    assert(d->curr);
    d->curr = d->curr->next;
}


static int got_arg(void *dsc)
{
    struct dsc *d = dsc;

    return !!d->curr;
}


static void check_arg(const struct dsc *d)
{
    if (!d->curr) {
	fprintf(stderr,"not enough arguments for format string\n");
	exit(1);
    }
}


static unsigned int get_unum(void *dsc)
{
    struct dsc *d = dsc;

    check_arg(d);
    if (!d->curr->value.type) d->curr->value = evaluate(d->curr->expr);
    return value_int(d->curr->value);
}


static int get_snum(void *dsc)
{
    struct dsc *d = dsc;
    int v,bits;

    v = get_unum(dsc);
    if (!d->curr->value.type) return 0;
    bits = d->curr->value.type->size*8;
    if (bits != sizeof(int)*8)
	if (v & (1 << (bits-1))) v |= ~0 << bits;
    return v;
}


static double get_float(void *dsc)
{
abort();
}


static const char *get_string(void *dsc)
{
    struct dsc *d = dsc;

    check_arg(d);
    if (!d->curr->value.type) d->curr->value = evaluate(d->curr->expr);
    return value_string(d->curr->value);
}


static void errorf(void *dsc,const char *msg,...)
{
    va_list ap;

    va_start(ap,msg);
    vfprintf(stderr,msg,ap);
    va_end(ap);
    exit(1);
}


static void warnf(void *dsc,const char *msg,...)
{
    va_list ap;

    va_start(ap,msg);
    fprintf(stderr,"warning: ");
    vfprintf(stderr,msg,ap);
    va_end(ap);
}


static struct sprintf_ops ops = {
    .first_arg = first_arg,
    .next_arg = next_arg,
    .got_arg = got_arg,
    .get_sint = get_snum,
    .get_uint = get_unum,
    .get_float = get_float,
    .get_string = get_string,
    .errorf = errorf,
    .warnf = warnf
};


void stmt_printf(const struct stmt *self)
{
    struct dsc dsc;
    struct arg_list *walk;
    const char *s;

    dsc.list = self->u.printf.args;
    for (walk = dsc.list; walk; walk = walk->next)
	walk->value.type = NULL;
    s = sprintf_generic(string_eval(self->u.printf.fmt),&ops,&dsc);
    printf("%s\n",s);
    free((char *) s);
}
